# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-10-01 03:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0023_process_entity_2'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='process',
            name='flow_collection',
        ),
    ]
