# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-09-06 15:23
from __future__ import unicode_literals

import django.contrib.postgres.fields.citext
from django.contrib.postgres.operations import CITextExtension
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0018_process_is_active'),
    ]

    operations = [
        CITextExtension(),
        migrations.AlterField(
            model_name='relation',
            name='category',
            field=django.contrib.postgres.fields.citext.CICharField(max_length=100),
        ),
    ]
